from . import BaseWidget
from ...core import EventLoop
from PySide2.QtGui import QCloseEvent
from PySide2.QtWidgets import QMainWindow
from ..interfaces import ClosedSignalInterface, ClassExecInterface


class MainWindow(QMainWindow, BaseWidget, ClosedSignalInterface, ClassExecInterface):
    def closeEvent(self, event: QCloseEvent):
        self.deal_with_close(event)

    def exec(self):
        with EventLoop() as event:
            self.show()
            self.closed.connect(event.quit)



import math
import typing
import operator
from . import SizeF
from PySide2.QtCore import QPoint, QPointF


class PointF(QPointF):
    def __init__(self, para: typing.Union[float, QPoint, QPointF], y: float = 0):
        if isinstance(para, QPoint) or isinstance(para, QPointF):
            super().__init__(para)
        else:
            super().__init__(para, y)

    @property
    def x(self) -> float:
        return super().x()

    @x.setter
    def x(self, value: float):
        self.setX(value)

    @property
    def y(self) -> float:
        return super().y()

    @y.setter
    def y(self, value: float):
        self.setY(value)

    @property
    def horizontal(self) -> 'PointF':
        return PointF(self.x, 0)

    @property
    def vertical(self) -> 'PointF':
        return PointF(0, self.y)

    def __calc__(self, other: typing.Union[int, float, 'PointF', SizeF], op: typing.Callable[[float, float], float]):
        if isinstance(other, PointF):
            return PointF(
                op(self.x, other.x),
                op(self.y, other.y)
            )
        elif isinstance(other, SizeF):
            return PointF(
                op(self.x, other.w),
                op(self.y, other.h)
            )
        else:
            return PointF(
                op(self.x, other),
                op(self.y, other)
            )

    def __add__(self, other: typing.Union[int, float, 'PointF', SizeF]):
        return self.__calc__(other, operator.add)

    __radd__ = __add__

    def __sub__(self, other: typing.Union[int, float, 'PointF', SizeF]):
        return self.__calc__(other, operator.sub)

    __rsub__ = None

    def __mul__(self, other: typing.Union[int, float, 'PointF', SizeF]):
        return self.__calc__(other, operator.mul)

    __rmul__ = __mul__

    def __truediv__(self, other: typing.Union[int, float, 'PointF', SizeF]):
        return self.__calc__(other, operator.truediv)

    def __neg__(self):
        return PointF(-self.x, -self.y)

    def __eq__(self, other: 'PointF'):
        return math.isclose(self.x, other.x) and math.isclose(self.y, other.y)



from . import Pen
from .utils import text_size
from ..core import PointF, RectF, Qt
from PySide2.QtGui import QPainter, QColor


class Painter(QPainter):  # pragma: no cover
    def draw_text(self, rect: RectF, text: str, color: QColor = None, align: Qt.TextAlignmentRole = Qt.AlignCenter):
        current_pen = self.pen()
        if color:
            self.setPen(Pen(color))

        self.drawText(rect, align, text)

        if color:
            self.setPen(current_pen)

    def draw_text_above(self, point: PointF, text, color: QColor = None, margin: float = 0.0):
        size = text_size(text=text, font=self.font())
        rect = RectF.bottom_center_at(point + PointF(0, -margin), size)
        self.draw_text(rect=rect, text=text, color=color)

    def draw_text_below(self, point: PointF, text, color: QColor = None, margin: float = 0.0):
        size = text_size(text=text, font=self.font())
        rect = RectF.top_center_at(point + PointF(0, margin), size)
        self.draw_text(rect=rect, text=text, color=color)

    def draw_text_left(self, point: PointF, text, color: QColor = None, margin: float = 0.0):
        size = text_size(text=text, font=self.font())
        rect = RectF.right_center_at(point + PointF(-margin, 0), size)
        self.draw_text(rect=rect, text=text, color=color)

    def draw_text_right(self, point: PointF, text, color: QColor = None, margin: float = 0.0):
        size = text_size(text=text, font=self.font())
        rect = RectF.left_center_at(point + PointF(margin, 0), size)
        self.draw_text(rect=rect, text=text, color=color)

    def draw_text_center(self, point: PointF, text, color: QColor = None):
        size = text_size(text=text, font=self.font())
        rect = RectF.center_at(point, size)
        self.draw_text(rect=rect, text=text, color=color)



import sys
from PySide2 import QtCore, QtWidgets


Qt = QtCore.Qt
# noinspection PyArgumentList
qt_app = QtWidgets.QApplication.instance()

if not qt_app:
    if getattr(sys, 'frozen', None):  # pragma: no cover
        # noinspection PyArgumentList
        QtWidgets.QApplication.addLibraryPath(QtCore.QDir.currentPath())

    qt_app = QtWidgets.QApplication([])



import typing
from .base import AbstractContainer
from .types import DictTypeMixin


class AbstractProject(AbstractContainer, DictTypeMixin):
    def get_value(self) -> dict:
        return self.attach(creator=dict)

    def set_value(self, value) -> None:
        self.value.clear()
        self.value.update(value)
        self.changed.emit(self.value)

    def get_property(self, name) -> typing.Any:
        return self.value.get(name)

    def set_property(self, name, value) -> None:
        self.value.update({name: value})
        self.changed.emit(self.value)



import typing
from ...model import base
from ..layouts import SquareLayout
from PySide2.QtWidgets import QWidget
from PySide2.QtWidgets import QDockWidget
from PySide2.QtWidgets import QMainWindow
from PySide2.QtWidgets import QHBoxLayout, QLayout


class BaseWidget(QWidget, base.AttachMixin):
    @property
    def center_widget(self) -> typing.Optional['BaseWidget']:
        return self.attach(lambda: None)

    @center_widget.setter
    def center_widget(self, widget: 'BaseWidget'):
        self.assign(widget)

    def set_central_widget(self, widget: 'BaseWidget', spacing: int=0, layout_type=QHBoxLayout):
        if self.center_widget is not None:
            self.layout().replaceWidget(self.center_widget, widget)
            self.center_widget.deleteLater()
            self.center_widget = widget
            return

        if isinstance(self, QMainWindow):
            self.setCentralWidget(widget)
        elif isinstance(self, QDockWidget):
            self.setWidget(widget)
        else:
            layout = layout_type()
            layout.setSpacing(spacing)
            layout.setContentsMargins(0, 0, 0, 0)
            layout.addWidget(widget)
            self.setLayout(layout)

        self.center_widget = widget

    def set_square_widget(self, widget: QWidget, spacing: int = 0):
        self.set_central_widget(widget, spacing, layout_type=SquareLayout)

    def set_focus(self):
        return self.setFocus()

    def set_enabled(self):
        return self.setEnabled(True)

    def set_disabled(self):
        return self.setEnabled(False)

    @property
    def enabled(self) -> bool:
        return self.isEnabled()

    @enabled.setter
    def enabled(self, value: bool):
        if value:
            self.set_enabled()
        else:
            self.set_disabled()



from ..base import ValueInterface
from ..base import ChangedValueEntity
from ..base import AbstractItem, AbstractProperty


class DictTypeMixin(ValueInterface):
    @property
    def value(self) -> dict:
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(value)


class DictTypeProperty(AbstractProperty, DictTypeMixin):
    pass


class DictPropertyInterface(ChangedValueEntity):
    @property
    def dict(self) -> DictTypeProperty:
        return self.attach(creator=DictTypeProperty, args=(
            self.get_dict_value, self.set_dict_value, self.changed
        ))

    def get_dict_value(self) -> dict:
        return self.get_value()

    def set_dict_value(self, value: dict):
        return self.set_value(value)


class DictItem(AbstractItem, DictPropertyInterface):
    pass



__name__ = 'quina'
__version__ = '0.0.1'
__description__ = 'PySide2 MVVM Framework'
__github__ = 'https://github.com/KD-Group/quina'

__author__ = 'SF-Zhou'
__email__ = 'sfzhou.scut@gmail.com'

from .core import SignalSender, connect_with
from .core import Qt, wait

from . import core
from . import model
from . import gui
from . import widgets



from .string_type import StringTypeMixin, StringTypeProperty, StringPropertyInterface, StringItem
from .integer_type import IntegerTypeMixin, IntegerTypeProperty, IntegerPropertyInterface, IntegerItem
from .float_type import FloatTypeMixin, FloatTypeProperty, FloatPropertyInterface, FloatItem
from .list_type import ListTypeMixin, ListTypeProperty, ListPropertyInterface, ListItem
from .dict_type import DictTypeMixin, DictTypeProperty, DictPropertyInterface, DictItem
from .string_list_type import StringListTypeMixin, StringListTypeProperty, StringListPropertyInterface, StringListItem



from . import BaseWidget
from PySide2.QtWidgets import QAction
from ..interfaces import ExcitedSignalInterface


class Action(QAction, BaseWidget, ExcitedSignalInterface):
    def set_excited_signal_connection(self):
        # noinspection PyUnresolvedReferences
        self.triggered.connect(self.excited.emit)

    def click(self):
        if self.enabled:
            self.trigger()



from . import SizeF, PointF
from PySide2.QtCore import QRectF


class RectF(QRectF):
    @staticmethod
    def left_top_at(point: PointF, size: SizeF) -> 'RectF':
        return RectF(point, size)

    @staticmethod
    def right_top_at(point: PointF, size: SizeF) -> 'RectF':
        return RectF(point - size.horizontal, size)

    @staticmethod
    def left_bottom_at(point: PointF, size: SizeF) -> 'RectF':
        return RectF(point - size.vertical, size)

    @staticmethod
    def right_bottom_at(point: PointF, size: SizeF) -> 'RectF':
        return RectF(point - size, size)

    @staticmethod
    def center_at(point: PointF, size: SizeF) -> 'RectF':
        return RectF(point - size / 2, size)

    @staticmethod
    def top_center_at(point: PointF, size: SizeF) -> 'RectF':
        return RectF(point - size.horizontal / 2, size)

    @staticmethod
    def bottom_center_at(point: PointF, size: SizeF) -> 'RectF':
        return RectF(point - size.horizontal / 2 - size.vertical, size)

    @staticmethod
    def left_center_at(point: PointF, size: SizeF) -> 'RectF':
        return RectF(point - size.vertical / 2, size)

    @staticmethod
    def right_center_at(point: PointF, size: SizeF) -> 'RectF':
        return RectF(point - size.horizontal - size.vertical / 2, size)

    @property
    def left_top(self) -> PointF:
        return PointF(self.topLeft())

    @property
    def right_top(self) -> PointF:
        return PointF(self.topRight())

    @property
    def left_bottom(self) -> PointF:
        return PointF(self.bottomLeft())

    @property
    def right_bottom(self) -> PointF:
        return PointF(self.bottomRight())

    @property
    def center(self) -> PointF:
        return PointF(super().center())

    @property
    def size(self) -> SizeF:
        return SizeF(self.width(), self.height())

    def __eq__(self, other: 'RectF') -> bool:
        return self.left_top == other.left_top and self.size == other.size



class ClassExecInterface:
    def __init__(self, *args, **kwargs):
        raise NotImplementedError

    def exec(self):
        raise NotImplementedError

    @classmethod
    def class_exec(cls, *args, **kwargs):
        cls(*args, **kwargs).exec()  # pragma: no cover



from .base_widget import BaseWidget

from .widget import Widget
from .dialog import Dialog
from .main_window import MainWindow
from .group_box import GroupBox
from .dock_widget import DockWidget

from .label import Label
from .line_edit import LineEdit

from .spin_box import SpinBox
from .double_spin_box import DoubleSpinBox

from .combo_box import ComboBox
from .list_widget import ListWidget

from .input_dialog import InputDialog

from .action import Action
from .shortcut import Shortcut
from .push_button import PushButton



import typing
from ..base import ValueInterface
from ..base import ChangedValueEntity
from ..base import AbstractItem, AbstractProperty


class StringListTypeMixin(ValueInterface):
    @property
    def value(self) -> typing.List[str]:
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(value)


class StringListTypeProperty(AbstractProperty, StringListTypeMixin):
    pass


class StringListPropertyInterface(ChangedValueEntity):
    @property
    def string_list(self) -> StringListTypeProperty:
        return self.attach(creator=StringListTypeProperty, args=(
            self.get_string_list_value, self.set_string_list_value, self.changed
        ))

    def get_string_list_value(self) -> typing.List[str]:
        return self.get_value()

    def set_string_list_value(self, value: typing.List[str]):
        return self.set_value(value)


class StringListItem(AbstractItem, StringListPropertyInterface):
    pass



from .qt_app import qt_app, Qt
from .signal_sender import SignalSender, connect_with
from .timer import Timer
from .event_loop import EventLoop, wait
from .size import SizeF
from .point import PointF
from .rect import RectF



__all__ = ('SignalSender', 'connect_with')
import typing
from PySide2.QtCore import QObject, Signal


class SignalWrapper(QObject):
    _signal = Signal(object)

    def emit_signal(self, value: tuple):
        # noinspection PyUnresolvedReferences
        self._signal.emit(value)

    def connect_slot(self, slot):
        # noinspection PyUnresolvedReferences
        self._signal.connect(slot)


class SignalSender:
    def __init__(self):
        self.signal = SignalWrapper()
        self.errors: typing.List[Exception] = []
        self.last_emit: typing.Tuple[list, dict] = None

    @property
    def last_error(self) -> Exception:
        return self.errors[-1] if len(self.errors) else None

    def emit(self, *args, **kwargs):
        self.errors.clear()

        self.last_emit = (args, kwargs)
        self.signal.emit_signal(self.last_emit)

        if self.last_error:
            raise self.last_error

    def emit_if_changed(self, *args, **kwargs):
        emit_value = (args, kwargs)
        if self.last_emit != emit_value:
            self.emit(*args, **kwargs)

    def connect(self, func, *args, **kwargs):
        errors = self.errors

        # never use `self` in slot function, which will cause a memory leak
        def slot_func(data, *_):
            try:
                return func(*args, *data[0], **kwargs, **data[1])
            except Exception as e:
                errors.append(e)
        slot_func.__name__ = getattr(func, '__name__', '<Lambda>')

        self.signal.connect_slot(slot_func)


def connect_with(signal: SignalSender, *args, **kwargs):
    def pack_func(func):
        signal.connect(func, *args, **kwargs)
        return func
    return pack_func



from . import BaseWidget
from ...model import WidgetStringItem
from PySide2.QtWidgets import QDoubleSpinBox


class DoubleSpinBox(QDoubleSpinBox, BaseWidget, WidgetStringItem):
    def get_string_value(self) -> str:
        return str(self.get_float_value())

    def set_string_value(self, value: str):
        self.set_float_value(float(value))

    def get_float_value(self) -> float:
        return self.value()

    def set_float_value(self, value: float):
        self.setValue(value)

    def set_changed_connection(self):
        # noinspection PyUnresolvedReferences
        self.valueChanged[float].connect(self.float.check_change)
        # noinspection PyUnresolvedReferences
        self.valueChanged[str].connect(self.string.check_change)



import quina
from setuptools import setup, find_packages


setup(
    name=quina.__name__,
    version=quina.__version__,
    description=quina.__description__,
    url=quina.__github__,
    author=quina.__author__,
    author_email=quina.__email__,

    license='GPL-3.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
    ],

    keywords='PySide2 MVVM',
    packages=find_packages(exclude=['docs', 'tests', 'examples']),
    install_requires=['PySide2', 'typing'],
    extras_require={'test': ['flake8', 'pytest', 'pytest-cov']}
)



# reference: http://www.qtcentre.org/threads/21329-keeping-a-widget-to-be-square
import typing
from PySide2.QtWidgets import QLayout, QLayoutItem
from PySide2.QtWidgets import QWidget, QWidgetItem
from PySide2.QtCore import Qt, QSize, QPoint, QRect


class SquareLayout(QLayout):
    def __init__(self, parent: QWidget = None,
                 default_item: QWidget = None):
        super().__init__(parent)

        self.item = default_item
        self.last_received_rect = QRect(0, 0, 0, 0)
        self.geometry_rect = QRect(0, 0, 0, 0)

        self.setSpacing(0)

    def addItem(self, item: QLayoutItem):
        if self.item is None:
            self.replace_item(item)

    def replace_item(self, item: QLayoutItem) -> typing.Optional[QLayoutItem]:
        old_item, self.item = self.item, item
        self.setGeometry(self.geometry_rect)
        return old_item

    def addWidget(self, w: QWidget):
        self.addItem(QWidgetItem(w))

    def replaceWidget(self, old_widget: QWidget, new_widget: QWidget):
        self.addWidget(new_widget)

    def take(self) -> QLayoutItem:
        item, self.item = self.item, None
        return item

    def takeAt(self, index: int) -> typing.Optional[QLayoutItem]:
        return self.take() if index == 0 else None

    def itemAt(self, index: int) -> typing.Optional[QLayoutItem]:
        return self.item if index == 0 else None

    def count(self):
        return 1 if self.item is not None else 0

    def expandingDirections(self):
        return Qt.Horizontal | Qt.Vertical

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        return self.item.minimumSize() if self.item is not None else QSize()

    def hasHeightForWidth(self):
        return False

    def geometry(self):
        return QRect(self.geometry_rect)

    def setGeometry(self, rect: QRect):
        if self.item is None or self.last_received_rect == rect:
            return

        self.last_received_rect = rect

        proper_size = self.calculate_proper_size(rect.size())
        proper_location = SquareLayout.calculate_center_location(rect.size(), proper_size)

        self.geometry_rect = QRect(proper_location, proper_size)
        self.item.setGeometry(self.geometry_rect)
        super().setGeometry(self.geometry_rect)

    def calculate_proper_size(self, from_size: QSize):
        minimum_length = min(from_size.width(), from_size.height())
        return QSize(minimum_length - self.spacing(), minimum_length - self.spacing())

    def update(self):
        self.setGeometry(self.geometry())

    @staticmethod
    def calculate_center_location(from_size: QSize, item_size: QSize):
        x = max(from_size.width() // 2 - item_size.width() // 2, 0)
        y = max(from_size.height() // 2 - item_size.height() // 2, 0)
        return QPoint(x, y)



from ..base import ValueInterface
from ..base import ChangedValueEntity
from ..base import AbstractItem, AbstractProperty


class ListTypeMixin(ValueInterface):
    @property
    def value(self) -> list:
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(value)


class ListTypeProperty(AbstractProperty, ListTypeMixin):
    pass


class ListPropertyInterface(ChangedValueEntity):
    @property
    def list(self) -> ListTypeProperty:
        return self.attach(creator=ListTypeProperty, args=(
            self.get_list_value, self.set_list_value, self.changed
        ))

    def get_list_value(self) -> list:
        return self.get_value()

    def set_list_value(self, value: list):
        return self.set_value(value)


class ListItem(AbstractItem, ListPropertyInterface):
    pass



from ..base import ValueInterface
from ..base import ChangedValueEntity
from ..base import AbstractItem, AbstractProperty


class StringTypeMixin(ValueInterface):
    @property
    def value(self) -> str:
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(value)


class StringTypeProperty(AbstractProperty, StringTypeMixin):
    pass


class StringPropertyInterface(ChangedValueEntity):
    @property
    def string(self) -> StringTypeProperty:
        return self.attach(creator=StringTypeProperty, args=(
            self.get_string_value, self.set_string_value, self.changed
        ))

    def get_string_value(self) -> str:
        return self.get_value()

    def set_string_value(self, value: str):
        return self.set_value(value)


class StringItem(AbstractItem, StringPropertyInterface):
    pass



from . import BaseWidget
from ...model import WidgetStringItem
from PySide2.QtWidgets import QLabel


class Label(QLabel, BaseWidget, WidgetStringItem):
    def get_value(self):
        return self.text()

    def set_value(self, value: str):
        self.setText(value)
        self.check_change()



from ..base import ValueInterface
from ..base import ChangedValueEntity
from ..base import AbstractItem, AbstractProperty


class FloatTypeMixin(ValueInterface):
    @property
    def value(self) -> float:
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(value)


class FloatTypeProperty(AbstractProperty, FloatTypeMixin):
    pass


class FloatPropertyInterface(ChangedValueEntity):
    @property
    def float(self) -> FloatTypeProperty:
        return self.attach(creator=FloatTypeProperty, args=(
            self.get_float_value, self.set_float_value, self.changed
        ))

    def get_float_value(self) -> float:
        return self.get_value()

    def set_float_value(self, value: float):
        return self.set_value(value)


class FloatItem(AbstractItem, FloatPropertyInterface):
    pass



from . import Timer
from PySide2.QtCore import QEventLoop


class EventLoop:
    def __init__(self, timeout_in_second: float=None):
        self.event = QEventLoop(None)
        self.timeout_in_second = timeout_in_second

    def quit(self):
        self.event.quit()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        if self.timeout_in_second is not None:
            t = Timer()
            t.timeout.connect(self.quit)
            t.start(seconds=self.timeout_in_second)
        self.event.exec_()


# non-blocking synchronous
def wait(seconds: float = 0.01):
    seconds = max(0.01, seconds)
    with EventLoop(timeout_in_second=seconds):
        pass
